/**
 * Ext GWT 3.0.0-beta4 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.widget.core.client.cell;

import com.google.gwt.cell.client.Cell;
import com.google.gwt.cell.client.Cell.Context;
import com.google.gwt.cell.client.ValueUpdater;
import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.Style.Position;
import com.google.gwt.editor.client.LeafValueEditor;
import com.google.gwt.event.logical.shared.ValueChangeEvent;
import com.google.gwt.event.logical.shared.ValueChangeHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.user.cellview.client.CellWidget;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.ui.HasValue;
import com.google.gwt.view.client.HasKeyProvider;
import com.google.gwt.view.client.ProvidesKey;
import com.sencha.gxt.cell.core.client.AbstractEventCell;
import com.sencha.gxt.cell.core.client.ResizableCell;
import com.sencha.gxt.core.client.GXT;
import com.sencha.gxt.core.client.resources.CommonStyles;
import com.sencha.gxt.core.client.util.DelayedTask;
import com.sencha.gxt.widget.core.client.Component;
import com.sencha.gxt.widget.core.client.ComponentHelper;

/**
 * A {@link Component} that wraps a {@link Cell}.
 * 
 * <p />
 * It is important to not that a single cell instance should not be used in a
 * single <code>CellComponent</code> instance.
 * 
 * @param <C> the type that the Cell represents
 */
public class CellComponent<C> extends Component implements HasKeyProvider<C>, HasValue<C>, LeafValueEditor<C> {

  /**
   * Create the default element used to wrap the Cell. The default element is a
   * div with display set to inline-block.
   * 
   * @return the default wrapper element
   */
  private static Element createDefaultWrapperElement() {
    Element div = Document.get().createDivElement();
    div.setClassName(CommonStyles.get().inlineBlock());
    // position relative is causing buttons in ie to not scroll with parent
    if (GXT.isIE6() || GXT.isIE7()) {
      div.getStyle().setPosition(Position.STATIC);
    }
    return div;
  }

  protected DelayedTask redrawTask = new DelayedTask() {

    @Override
    public void onExecute() {
      redraw();
    }
  };

  /**
   * The cell being wrapped.
   */
  private final Cell<C> cell;

  /**
   * The key provider for the value.
   */
  private ProvidesKey<C> keyProvider;

  /**
   * The current cell value.
   */
  private C value;

  /**
   * The {@link ValueUpdater} used to trigger value update events.
   */
  protected final ValueUpdater<C> valueUpdater = new ValueUpdater<C>() {
    public void update(C value) {
      // no need to redraw, the Cell took care of it
      setValue(value, true, false);
    }
  };

  /**
   * Construct a new {@link CellWidget} with the specified cell and an initial
   * value of <code>null</code>.
   * 
   * @param cell the cell to wrap
   */
  protected CellComponent(Cell<C> cell) {
    this(cell, null, null);
  }

  protected CellComponent(Cell<C> cell, boolean deferSetValue) {
    this.cell = cell;
    this.keyProvider = null;
    setElement(createDefaultWrapperElement());
    getElement().repaint();
    CellWidgetImplHelper.sinkEvents(this, cell.getConsumedEvents());
  }

  /**
   * Construct a new {@link CellWidget} with the specified cell and initial
   * value.
   * 
   * @param cell the cell to wrap
   * @param initialValue the initial value of the Cell
   */
  protected CellComponent(Cell<C> cell, C initialValue) {
    this(cell, initialValue, null);
  }

  /**
   * Construct a new {@link CellWidget} with the specified cell, initial value,
   * and key provider.
   * 
   * @param cell the cell to wrap
   * @param initialValue the initial value of the Cell
   * @param keyProvider the key provider used to get keys from values
   */
  protected CellComponent(Cell<C> cell, C initialValue, ProvidesKey<C> keyProvider) {
    this(cell, initialValue, keyProvider, createDefaultWrapperElement());
  }

  /**
   * Creates a {@link CellWidget} with the specified cell, initial value, key
   * provider, using the specified element as the wrapper around the cell.
   * 
   * @param cell the cell to wrap
   * @param initialValue the initial value of the Cell
   * @param keyProvider the key provider used to get keys from values
   * @param elem the browser element to use
   */
  protected CellComponent(Cell<C> cell, C initialValue, ProvidesKey<C> keyProvider, Element elem) {
    this.cell = cell;
    this.keyProvider = keyProvider;
    setElement(elem);
    CellWidgetImplHelper.sinkEvents(this, cell.getConsumedEvents());
    setValue(initialValue);
  }

  /**
   * Construct a new {@link CellWidget} with the specified cell and key
   * provider, and an initial value of <code>null</code>.
   * 
   * @param cell the cell to wrap
   * @param keyProvider the key provider used to get keys from values
   */
  protected CellComponent(Cell<C> cell, ProvidesKey<C> keyProvider) {
    this(cell, null, keyProvider);
  }

  public HandlerRegistration addValueChangeHandler(ValueChangeHandler<C> handler) {
    return addHandler(handler, ValueChangeEvent.getType());
  }

  /**
   * Get the {@link Cell} wrapped by this widget.
   * 
   * @return the {@link Cell} being wrapped
   */
  public Cell<C> getCell() {
    return cell;
  }

  public ProvidesKey<C> getKeyProvider() {
    return keyProvider;
  }

  public C getValue() {
    return value;
  }

  @Override
  public void onBrowserEvent(Event event) {
    CellWidgetImplHelper.onBrowserEvent(this, event);
    super.onBrowserEvent(event);

    // Forward the event to the cell.
    String eventType = event.getType();
    if (cell.getConsumedEvents().contains(eventType)) {
      cell.onBrowserEvent(createContext(), getElement(), value, event, valueUpdater);
    }
  }

  /**
   * Redraw the widget.
   */
  public void redraw() {
    // boolean editing = getCell().isEditing(createContext(), getElement(),
    // value);
    SafeHtmlBuilder sb = new SafeHtmlBuilder();
    cell.render(createContext(), value, sb);
    getElement().setInnerHTML(sb.toSafeHtml().asString());

    // if (editing) {
    // cell.resetFocus(createContext(), getElement(), value);
    // }

    /*
     * The rendered Cell should fill the root element so height and width styles
     * applied to the widget also apply to the Cell. If there is exactly one
     * child element, the height and width are set to 100% to fill the parent.
     */
    // Element child = getElement().getFirstChildElement();
    // if (child != null && child.getNextSiblingElement() == null) {
    // child.getStyle().setHeight(100, Unit.PCT);
    // child.getStyle().setWidth(100, Unit.PCT);
    // }
  }

  /**
   * {@inheritDoc}
   * <p>
   * This method will redraw the widget if the new value does not equal the
   * existing value.
   * </p>
   */
  public void setValue(C value) {
    setValue(value, false, true);
  }

  /**
   * {@inheritDoc}
   * <p>
   * This method will redraw the widget if the new value does not equal the
   * existing value.
   * </p>
   */
  public void setValue(C value, boolean fireEvents) {
    setValue(value, fireEvents, true);
  }

  /**
   * Sets this object's value and optionally redraw the widget. Fires
   * {@link com.google.gwt.event.logical.shared.ValueChangeEvent} when
   * fireEvents is true and the new value does not equal the existing value.
   * Redraws the widget when redraw is true and the new value does not equal the
   * existing value.
   * 
   * @param value the object's new value
   * @param fireEvents fire events if true and value is new
   * @param redraw redraw the widget if true and value is new
   */
  public void setValue(C value, boolean fireEvents, boolean redraw) {
    C oldValue = getValue();
    if (value != oldValue && (oldValue == null || !oldValue.equals(value))) {
      this.value = value;
      if (redraw) {
        redraw();
      }
      if (fireEvents) {
        ValueChangeEvent.fire(this, value);
      }
    }
  }

  /**
   * Get the {@link Context} for the cell.
   */
  protected Context createContext() {
    return new DefaultHandlerManagerContext(0, 0, getKey(value), ComponentHelper.ensureHandlers(this));
  }

  /**
   * Get the key for the specified value.
   * 
   * @param value the value
   * @return the key
   */
  protected Object getKey(C value) {
    // since cells can't be reused with CellComponent we use a hard coded key
    // that will remain consistent
    // unlike the GWT impl which uses the actual value when key provider is
    // null. also, the GWT cell code
    // does not handle null keys
    Object key = (keyProvider == null) ? "component" : keyProvider.getKey(value);
    return key;
  }

  @Override
  protected void onDisable() {
    super.onDisable();
    if (cell instanceof AbstractEventCell<?>) {
      ((AbstractEventCell<?>) cell).setDisableEvents(true);
    }
  }

  @Override
  protected void onEnable() {
    super.onEnable();
    if (cell instanceof AbstractEventCell<?>) {
      ((AbstractEventCell<?>) cell).setDisableEvents(false);
    }
  }

  @Override
  protected void onResize(int width, int height) {
    super.onResize(width, height);
    if (cell instanceof ResizableCell) {
      ResizableCell rc = (ResizableCell) cell;
      if (rc.redrawOnResize()) {
        redraw();
      } else {
        rc.setSize(getElement(), width, height);
      }
    }
  }

  protected void setKeyProvider(ProvidesKey<C> keyProvider) {
    this.keyProvider = keyProvider;
  }
}